#ifndef POISSON_HEADER
#define POISSON_HEADER

#include <cstdlib>
#include <omp.h>
#include <cilk/cilk.h>
#include <cilk/reducer_max.h>
#include <cilk/holder.h>
#include <cilk/cilk_api.h>
#include <cmath>
#include <iostream>

const double precision = 0.1;

class Mesh
{
protected:
    double *x, *y;
    int _meshSize;
    double _cMin, _cMax;
    double _step;

    void _initMesh() {
        x = new double[_meshSize+2];    // include boundary points
        y = new double[_meshSize+2];
        _step = (_cMax - _cMin) / double(_meshSize + 1);
        for(int i=0; i<_meshSize; i++) {
            x[i] = _cMin + double(i) * _step;
            y[i] = _cMin + double(i) * _step;
        }
    }

    void _destroyMesh() {
        delete [] x;
        delete [] y;
    }
    
public:
    Mesh(int meshSize, double cMin, double cMax)
    {
        if (cMin > cMax) {
            _cMin = cMax;
            _cMax = cMin;
        } else {
            _cMin = cMin;
            _cMax = cMax;
        }
        _meshSize = meshSize;
        _initMesh();
    }
    virtual ~Mesh()
    {
        _destroyMesh();
    }

    int GetMeshSize() {
        return _meshSize;
    }
    double Step() {
        return _step;
    }
};

class Problem : public Mesh 
{
protected:
    double **u; // solution
    double **f; // mesh function
    double *_getRand(double min, double max) {
        double *data = new double[_meshSize+2]; // include boundary points
        for(int i=0; i<_meshSize+2; i++) {
    		data[i] = min+((double)rand()/(double)RAND_MAX)*(max-min);            
        }
        return data;
    }
    void _initRand(double min, double max) {
        u = new double*[_meshSize+2];
        for(int i=0; i<_meshSize+2; i++) {
            u[i] = _getRand(min, max);
        }
    }
    
    // override this function in derived class to provide other boundary conditions
    virtual void _initBoundary() {      
      	for(int i = 1; i < _meshSize+1; i++){
    		u[0][i]   =  100 - 200*x[i];
    		u[i][0]   =  100 - 200*y[i];
    		u[_meshSize+1][i] = -100 + 200*x[i];
    		u[i][_meshSize+1] = -100 + 200*y[i];
    	}	      
    }

    // override this function in derived class to provide other right-hand side of Poisson eq.
    virtual void _initMeshFunction() {  
        f = new double*[_meshSize+2];
    	for(int i = 0; i < _meshSize+2; i++){
            f[i] = new double[_meshSize+2];
    		for(int j = 0; j < _meshSize+2; j++){
    			f[i][j] = 0;
    		}
    	}        
    }

    void _destroySolution() {
        for(int i=0; i<_meshSize+2; i++) 
            delete [] u[i];
        delete [] u;
    }

    void _destroyMeshFunction() {
        for(int i=0; i<_meshSize+2; i++)
			delete [] f[i];
		delete [] f;
    }	
public:
    Problem(int meshSize, double cMin, double cMax)
        : Mesh(meshSize, cMin, cMax) {
        srand(0);  
        _initRand(-100.0, 100.0);
        _initBoundary();  
        _initMeshFunction();
    }   
    virtual ~Problem() {
        _destroySolution();
		_destroyMeshFunction();
    }

    void ReinitProblem() {
        _destroySolution();
        srand(0);  
        _initRand(-100.0, 100.0);
        _initBoundary();
    }
    
    
    double ** U() {
        return u;
    }
    
    double ** F() {
        return f;
    } 
};


// basic class for all solvers, abstact (virtual) class
class Solver
{
protected:
    Problem *p;
public:
    virtual ~Solver() {}
    void SetProblem(Problem *problem) {
        p = problem;
    }

    double RunAndMeasure() {
    	double time = omp_get_wtime();
    	Solve();
    	time = omp_get_wtime() - time;
        return time;
    }
    
    virtual void Solve() = 0;   // pure virtual function, MUST be overriden in derived classes to provide functionality
};

// Serial solver class (Gauss-Seidel method)
class SerialSeidelSolver : public Solver {
public:
    virtual void Solve() {
        if (p == 0) return; // no problem set
        int N = p->GetMeshSize();
        double variation(precision), oldValue, diff;
        double step = p->Step(); // get step for the problem

        double **u = p->U();
        double **f = p->F();

        int count = 0;
        while(variation >= precision) {
            count++;
            variation = 0;
            for(int i=1; i<N; i++) {
                for(int j=1; j<N; j++) {
                    oldValue = u[i][j];
                    u[i][j] = 
                        0.25 * 
                        (u[i-1][j] + u[i+1][j] 
                        +u[i][j-1] + u[i][j+1] - 
                        step * step * f[i][j]);
                    diff = abs(oldValue-u[i][j]);
                    if (variation < diff) variation = diff;
                }
            }
//            std::cout << variation << std::endl;
        }
//        std::cout << count << " iterations" << std::endl;
    }
};

// Serial solver class (Gauss-Jakobi method)
class SerialJakobiSolver : public Solver {
public:
    virtual void Solve() {
        if (p == 0) return; // no problem set
        // prepare temporary storage for solution
        int N = p->GetMeshSize();
        double **u_t = new double*[N+2];
        for(int i=0; i<N+2; i++)
            u_t[i] = new double[N+2];

        double **u = p->U();
        double **f = p->F();

        
        double variation(precision), oldValue, diff;
        double step = p->Step(); // get step for the problem
        int count = 0;
        while(variation >= precision) {
            count ++;
            variation = 0;
            for(int i=1; i<N; i++) {
                for(int j=1; j<N; j++) {
                    oldValue = u[i][j];
                    u_t[i][j] = 
                        0.25 * 
                        (u[i-1][j] + u[i+1][j]
                        + u[i][j-1] + u[i][j+1] - 
                        step * step * f[i][j]);
                    diff = abs(oldValue - u_t[i][j]);
                    if (variation < diff) variation = diff;
                }
            }
            // copy solution from temporary storage
            for(int i=1; i<N; i++)
                for(int j=1; j<N; j++)
                    u[i][j] = u_t[i][j];
//            std::cout << variation << std::endl;
        }
//        std::cout << count << " iterations" << std::endl;
    }
};


// parallel Gauss-Seidel solver using cilk_for
class CilkSeidelSolver : public Solver {
public:
    virtual void Solve() {
        if (p == 0) return; // no problem set
        const int N = p->GetMeshSize();
        double variation(precision), oldValue, diff;
        double step = p->Step(); // get step for the problem
        int count = 0;

        double **u = p->U();
        double **f = p->F();

        while(variation >= precision) {
            count++;
            variation = 0;
            cilk::reducer_max<double> reducer(0);
            cilk_for(int i=1; i<N; i++) {
                double vm;
                vm = 0;
                for(int j=1; j<N; j++) {
                    oldValue = u[i][j];
                    u[i][j] = 
                        0.25 * 
                        (u[i-1][j] + u[i+1][j]
                        + u[i][j-1] + u[i][j+1] - 
                        step * step * f[i][j]);
                    diff = abs(oldValue-u[i][j]);
                    if (vm < diff) vm = diff;
                }
            	reducer.calc_max(vm);
            }
            variation = reducer.get_value();
//            std::cout << variation << std::endl;
        }
//        std::cout << count << " iterations" << std::endl;
    }
};

// parallel Gauss-Seidel solver using OpenMP
class OpenMPSeidelSolver : public Solver {
public:
    virtual void Solve() {
        if (p == 0) return; // no problem set
        const int N = p->GetMeshSize();
        double variation(precision), oldValue, diff;
        double step = p->Step(); // get step for the problem
        int count = 0;

        double **u = p->U();
        double **f = p->F();

        double vm;

        int i,j;
        
        omp_lock_t variation_lock;
        omp_init_lock(&variation_lock);
    
        while(variation >= precision) {
            count++;
            variation = 0;
#pragma omp parallel for shared(u,f,N,variation) private(i,j,oldValue,vm,diff)
            for(i=1; i<N; i++) {
                vm = 0;
                for(j=1; j<N; j++) {
                    oldValue = u[i][j];
                    u[i][j] = 
                        0.25 * 
                        (u[i-1][j] + u[i+1][j]
                        + u[i][j-1] + u[i][j+1] - 
                        step * step * f[i][j]);
                    diff = abs(oldValue-u[i][j]);
                    if (vm < diff) vm = diff;
                }
                omp_set_lock(&variation_lock);
                if (variation < vm) variation = vm;
                omp_unset_lock(&variation_lock);
            }
          //  std::cout << variation << std::endl;
        }
//        std::cout << count << " iterations" << std::endl;
    }
};

// Parallel solver class (Gauss-Jakobi method) using OpenMP
class OpenMPJakobiSolver : public Solver {
public:
    virtual void Solve() {
        if (p == 0) return; // no problem set
        // prepare temporary storage for solution
        int N = p->GetMeshSize();
        double **u_t = new double*[N+2];
        for(int i=0; i<N+2; i++)
            u_t[i] = new double[N+2];

        double **u = p->U();
        double **f = p->F();

        omp_lock_t variation_lock;
        omp_init_lock(&variation_lock);
        
        double variation(precision), oldValue, diff;
        double step = p->Step(); // get step for the problem
        int count = 0;
        double vm;
		int i,j;
        while(variation >= precision) {
            count ++;
            variation = 0;
#pragma omp parallel for shared(u,u_t,f,N,variation) private(i,j,oldValue,vm,diff)
            for(i=1; i<N; i++) {
                vm = 0;
                for(j=1; j<N; j++) {
                    oldValue = u[i][j];
                    u_t[i][j] = 
                        0.25 * 
                        (u[i-1][j] + u[i+1][j]
                        + u[i][j-1] + u[i][j+1] - 
                        step * step * f[i][j]);
                    diff = abs(oldValue - u_t[i][j]);
                    if (vm < diff) vm = diff;
                }
                omp_set_lock(&variation_lock);
                if (variation < vm) variation = vm;
                omp_unset_lock(&variation_lock);
            }
            // copy solution from temporary storage
#pragma omp parallel for shared(u,u_t,N) private(i,j)
            for(i=1; i<N; i++)
                for(j=1; j<N; j++)
                    u[i][j] = u_t[i][j];
//            std::cout << variation << std::endl;
        }
//        std::cout << count << " iterations" << std::endl;
    }
};

class CilkJakobiSolver : public Solver {
public:
    virtual void Solve() {
        if (p == 0) return; // no problem set
        // prepare temporary storage for solution
        int N = p->GetMeshSize();
        double **u_t = new double*[N+2];
        for(int i=0; i<N+2; i++)
            u_t[i] = new double[N+2];

        double **u = p->U();
        double **f = p->F();

        
        double variation(precision), oldValue, diff;
        double step = p->Step(); // get step for the problem
        int count = 0;
        while(variation >= precision) {
            count ++;
            variation = 0;
            cilk::reducer_max<double> reducer(0);
            reducer.set_value(0);
            cilk_for(int i=1; i<N; i++) {
                double vm;
                vm = 0;
                for(int j=1; j<N; j++) {
                    oldValue = u[i][j];
                    u_t[i][j] = 
                        0.25 * 
                        (u[i-1][j] + u[i+1][j]
                        + u[i][j-1] + u[i][j+1] - 
                        step * step * f[i][j]);
                    diff = abs(oldValue - u_t[i][j]);
                    if (vm < diff) vm = diff;
                }
                reducer.calc_max(vm);
            }
            variation = reducer.get_value();
            // copy solution from temporary storage
            for(int i=1; i<N; i++)
               for(int j=1; j<N; j++)
                    u[i][j] = u_t[i][j];
//            std::cout << variation << std::endl;
        }
//        std::cout << count << " iterations" << std::endl;
    }
};

class OpenMPWaveSolver : public Solver {
public:
    virtual void Solve() {
	if (p == 0) return;
	double variation,d,old;
	int N = p->GetMeshSize();
	double *vm = new double[N+3];
	
	double **u = p->U();
	double **f = p->F();
	
	int i,j,nx;
	
	omp_set_num_threads(4);
	omp_lock_t variation_max_lock;
	omp_init_lock(&variation_max_lock);
	variation = precision;
	
	double step = p->Step();
	
	while(variation >= precision) {
	    variation = 0;
	    for(nx=1; nx<N+1; nx++) {
		vm[nx]=0;
#pragma omp parallel for shared(u,N,nx,vm) private(i,j,old,d)
		for(i=1; i<nx+1; i++) {
		    j = nx + 1 - i;
		    old = u[i][j];
		    u[i][j] = 0.25 * (u[i-1][j] + u[i+1][j]
				+ u[i][j-1] + u[i][j+1] 
				- step * step * f[i][j]);
		    d = abs(old - u[i][j]);
		    if (vm[i] < d) vm[i] = d;
		}
	    }
	    for(nx=N-1; nx>0; nx--) {
#pragma omp parallel for shared(u,N,nx,vm) private(i,j,old,d)
		for(i=N-nx+1; i<N+1; i++) {
		    j = 2*N - nx - i + 1;
		    old = u[i][j];
		    u[i][j] = 0.25 * (u[i-1][j] + u[i+1][j]
				    + u[i][j-1] + u[i][j+1]
				    - step * step * f[i][j]);
		    d = abs(old - u[i][j]);
		    if (vm[i] < d) vm[i] = d;
		}
	    }
#pragma omp parallel for shared(nx,vm,variation) private(i)
	    for(i=1; i<nx+1; i++) {
		omp_set_lock(&variation_max_lock);
		if (variation < vm[i]) variation = vm[i];
		omp_unset_lock(&variation_max_lock);
	    }
	}
	delete [] vm;
    }
};

class CilkWaveSolver : public Solver {
public:
    virtual void Solve() {
	if (p == 0) return;
	double variation,d,old;
	int N = p->GetMeshSize();
	double *vm = new double[N+3];
	
	double **u = p->U();
	double **f = p->F();
	
	int j,nx;
	
	variation = precision;
	
	double step = p->Step();
	
	while(variation >= precision) {
	    variation = 0;
	    for(nx=1; nx<N+1; nx++) {
		vm[nx]=0;
			cilk_for(int i=1; i<nx+1; i++) {
			    j = nx + 1 - i;
			    old = u[i][j];
			    u[i][j] = 0.25 * (u[i-1][j] + u[i+1][j]
					+ u[i][j-1] + u[i][j+1] 
					- step * step * f[i][j]);
			    d = abs(old - u[i][j]);
			    if (vm[i] < d) vm[i] = d;
			}
	    }
	    for(nx=N-1; nx>0; nx--) {
			cilk_for(int i=N-nx+1; i<N+1; i++) {
			    j = 2*N - nx - i + 1;
			    old = u[i][j];
			    u[i][j] = 0.25 * (u[i-1][j] + u[i+1][j]
					    + u[i][j-1] + u[i][j+1]
					    - step * step * f[i][j]);
			    d = abs(old - u[i][j]);
			    if (vm[i] < d) vm[i] = d;
			}
	    }
	    for(int i=1; i<nx+1; i++) {
			if (variation < vm[i]) variation = vm[i];
	    }
	}
	delete [] vm;
    }
};


#endif
